/* Copyright (c) 1997-2004
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
   http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ifndef _POLYMAKE_NOTHING_H
#define _POLYMAKE_NOTHING_H "$Project: polymake $$Id: nothing.h,v 1.16 2004/06/22 16:22:42 gawrilow Exp $"

#if defined(__GNUC__) && __GNUC_MINOR__<3 && !defined(__APPLE__) && !defined(__COMO__)
#pragma interface
#endif

#ifndef _POLYMAKE_TYPE_MANIP_H
# include <type_manip.h>
#endif

namespace pm {

/// Structure denoting the absence of data
struct nothing {
   void operator= (const nothing&) const { }
   bool operator== (const nothing&) const { return true; }
   bool operator!= (const nothing&) const { return false; }
   bool operator< (const nothing&) const { return false; }
   bool operator> (const nothing&) const { return false; }
   bool operator<= (const nothing&) const { return true; }
   bool operator>= (const nothing&) const { return true; }
};

inline void relocate(const nothing*, const nothing*) { }

template <>
struct spec_object_traits<nothing>
   : spec_object_traits< int2type<object_classifier::unspecified> > {
   typedef nothing model;	// prevent mixing this with any other type
};

/// Masquerade class for suppressing I/O and conversion
template <typename T, typename Expected=nothing>
struct ignore { };

template <typename T, typename Expected>
struct spec_object_traits< ignore<T,Expected> >
   : generic_object_traits<T> { };

template <typename T>
struct unveil_ignore {
   typedef T type;
   typedef void expected_type;
};

template <typename T, typename Expected>
struct unveil_ignore< ignore<T,Expected> > {
   typedef T type;
   typedef Expected expected_type;
};

} // end namespace pm

namespace polymake {
   using pm::nothing;
}

namespace std {
   /// Do nothing
   inline void swap(const pm::nothing&, const pm::nothing&) { }
}

#endif // _POLYMAKE_NOTHING_H

// Local Variables:
// mode:C++
// End:
