#  Copyright (c) 1997-2004
#  Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
#  http://www.math.tu-berlin.de/polymake,  mailto:polymake@math.tu-berlin.de
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2, or (at your option) any
#  later version: http://www.gnu.org/licenses/gpl.txt.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#----------------------------------------------------------------------------
#  $Project: polymake $$Id: Poly.pm,v 1.3 2004/10/26 19:02:51 gawrilow Exp $

use strict;

package RefHash;
use Poly::Ext;

package Poly;
use Poly::Ext;

####################################################################################

sub min($$) {  $_[0]<$_[1] ? $_[0] : $_[1]  }
sub max($$) {  $_[0]>$_[1] ? $_[0] : $_[1]  }
sub assign_min($$) { $_[0] > $_[1] and $_[0]=$_[1] }
sub assign_max($$) { $_[0] < $_[1] and $_[0]=$_[1] }
sub assign_min_max($$$) { $_[0] > $_[2] ? ($_[0]=$_[2]) : $_[1] < $_[2] && ($_[1]=$_[2]) }
####################################################################################
sub contains($$) {		# \@list, scalar => bool
   my ($list, $item)=@_;
   foreach my $elem (@$list) {
      ref($elem) ? contains($elem,$item) : $elem==$item
      and return 1;
   }
   return 0;
}
####################################################################################
sub uniq {
   my %seen;
   grep { !($seen{$_}++) } @_;
}
####################################################################################
sub sorted_uniq {
   my $i=-1;
   grep { ++$i == 0 || $_[$i-1] ne $_ } @_;
}
####################################################################################
sub num_sorted_uniq {
   my $i=-1;
   grep { ++$i == 0 || $_[$i-1] != $_ } @_;
}
####################################################################################
# \@list, scalar, [order] => bool
# @list must be sorted in increasing order (or decreasing if $order==-1)
sub binsearch($$;$) {
   my ($list, $item, $order)=(@_, 1);
   my ($l, $h)=(0, scalar @$list);
   while ($l<$h) {
      my $m=($l+$h)/2;
      my $cmp=$item <=> $list->[$m];
      return 1 if !$cmp;
      $cmp!=$order ? ($h=$m) : ($l=$m+1);
   }
   return 0;
}
####################################################################################
sub deep_copy($) {
   my ($src)=@_;
   !ref($src) || is_object($src)
     ? $src
   : is_hash($src)
     ? { map { $_ => deep_copy($src->{$_}) } keys %$src }
   : is_ARRAY($src)
     ? [ map { deep_copy($_) } @$src ]
     : $src
}
####################################################################################
sub read_block($) {		# \FILEHANDLE => \@data, \@comments
   my ($in)=@_;
   my (@data, @comments);

   while (<$in>) {
      s/^\s+//;
      last if !length;
      $_.="\n" if substr($_,-1,1) ne "\n";
      if (substr($_,0,1) eq '#') {
	 push @comments, $_;
      } else {
	 push @data, $_;
      }
   }
   if (@data) {
      # for the rare case of the last file line lacking \n
      $data[-1].="\n" if substr($data[-1],-1,1) ne "\n";
   }
   wantarray ? (\@data, \@comments) : \@data;
}
####################################################################################
sub read_block_chomp($) {		# \FILEHANDLE => \@data
   my $data=&read_block;
   foreach (@$data) { chomp };
   return $data;
}
####################################################################################
our $dbg_prefix="polymake: ";
our $err_prefix=$dbg_prefix;

{ my $sel=select; select STDERR; $|=1; select $sel; }

sub dbg_print {
   # FIXME: should be connected to the shell output facilities
   print STDERR $dbg_prefix, @_, "\n";
}
####################################################################################
sub err_print {
   # FIXME: should be connected to the shell output facilities
   print STDERR $err_prefix, "ERROR: ", @_, "\n";
}
####################################################################################
sub warn_print {
   # FIXME: should be connected to the shell output facilities
   print STDERR $err_prefix, "WARNING: ", @_, "\n";
}
####################################################################################
sub process_eval_errors($) {
   my ($err)=@_;
   my $text="";
   while ($err =~ /^(.*?) at (.*?)(?:\[.*?\])? (line \d+)(?:,(.*)$)?/msg) {
      $text .= "\"$2\", $3: $1.$4\n";
   }
   $text;
}
####################################################################################
sub print_subs {
   no strict 'refs';
   my ($pkg)=@_;  $pkg ||= caller;
   print STDERR "package $pkg\n";
   foreach my $k (keys %{"$pkg\::"}) {
      if (defined *{"$pkg\::$k"}{CODE}) {
	 print STDERR "$k\n";
      }
   }
   foreach (@{"$pkg\::ISA"}) {
      print_subs($_);
   }
}
####################################################################################
declare_lvalue(\&retrieve);

*UNIVERSAL::can=\&can;

sub is_array {
   is_object($_[0]) && can($_[0],".constructor") ? defined(overload::Method($_[0],'@{}')) : is_ARRAY($_[0])
}

sub croak { namespaces::disable(); goto &Carp::croak }
sub confess { namespaces::disable(); goto &Carp::confess }

1
